use std::pin::Pin;

use super::{util::TaggedAtomicPtr, *};

fn pin_ptr<T>(pin: &Pin<&mut T>) -> NonNull<T> {
    let a = &raw const **pin;
    unsafe { NonNull::new_unchecked(a.cast_mut()) }
}

#[test]
fn job_list_pop_back() {
    let mut list = JobList::new();
    let mut a = pin!(Job::empty());
    let mut b = pin!(Job::empty());
    let mut c = pin!(Job::empty());

    unsafe {
        list.push_front(a.as_ref());
        list.push_front(b.as_ref());
        list.push_back(c.as_ref());
    }

    assert_eq!(list.pop_back(), Some(pin_ptr(&c)));
    unsafe {
        list.push_front(c.as_ref());
    }
    assert_eq!(list.pop_back(), Some(pin_ptr(&a)));
    assert_eq!(list.pop_back(), Some(pin_ptr(&b)));
    assert_eq!(list.pop_back(), Some(pin_ptr(&c)));
    assert_eq!(list.pop_back(), None);
    assert_eq!(list.pop_front(), None);
}

#[test]
fn job_list_pop_front() {
    let mut list = JobList::new();
    let mut a = pin!(Job::<()>::empty());
    let mut b = pin!(Job::<()>::empty());
    let mut c = pin!(Job::<()>::empty());

    unsafe {
        list.push_front(a.as_ref());
        list.push_front(b.as_ref());
        list.push_back(c.as_ref());
    }

    assert_eq!(list.pop_front(), Some(pin_ptr(&b)));
    unsafe {
        list.push_back(b.as_ref());
    }
    assert_eq!(list.pop_front(), Some(pin_ptr(&a)));
    assert_eq!(list.pop_front(), Some(pin_ptr(&c)));
    assert_eq!(list.pop_front(), Some(pin_ptr(&b)));
    assert_eq!(list.pop_front(), None);
    assert_eq!(list.pop_back(), None);
}

#[test]
fn unlink_job_middle() {
    let mut list = JobList::new();
    let mut a = pin!(Job::<()>::empty());
    let mut b = pin!(Job::<()>::empty());
    let mut c = pin!(Job::<()>::empty());

    unsafe {
        list.push_front(a.as_ref());
        list.push_front(b.as_ref());
        list.push_front(c.as_ref());
    }

    unsafe {
        b.unlink();
    }

    assert_eq!(list.pop_front(), Some(pin_ptr(&c)));
    assert_eq!(list.pop_front(), Some(pin_ptr(&a)));
    assert_eq!(list.pop_front(), None);
    assert_eq!(list.pop_back(), None);
}

#[test]
fn unlink_job_front() {
    let mut list = JobList::new();
    let mut a = pin!(Job::<()>::empty());
    let mut b = pin!(Job::<()>::empty());
    let mut c = pin!(Job::<()>::empty());

    unsafe {
        list.push_front(a.as_ref());
        list.push_front(b.as_ref());
        list.push_front(c.as_ref());
    }

    unsafe {
        a.unlink();
    }

    assert_eq!(list.pop_front(), Some(pin_ptr(&c)));
    assert_eq!(list.pop_front(), Some(pin_ptr(&b)));
    assert_eq!(list.pop_front(), None);
    assert_eq!(list.pop_back(), None);
}

#[test]
fn unlink_job_back() {
    let mut list = JobList::new();
    let mut a = pin!(Job::<()>::empty());
    let mut b = pin!(Job::<()>::empty());
    let mut c = pin!(Job::<()>::empty());

    unsafe {
        list.push_front(a.as_ref());
        list.push_front(b.as_ref());
        list.push_front(c.as_ref());
    }

    unsafe {
        c.unlink();
    }

    assert_eq!(list.pop_front(), Some(pin_ptr(&b)));
    assert_eq!(list.pop_front(), Some(pin_ptr(&a)));
    assert_eq!(list.pop_front(), None);
    assert_eq!(list.pop_back(), None);
}

#[test]
fn unlink_job_single() {
    let mut list = JobList::new();
    let a = pin!(Job::<()>::empty());

    unsafe {
        list.push_front(a.as_ref());
    }

    unsafe {
        a.unlink();
    }

    assert_eq!(list.pop_front(), None);
    assert_eq!(list.pop_back(), None);
}

#[test]
fn tagged_ptr_exchange() {
    let boxed = Box::into_raw(Box::new(42usize));
    let ptr = TaggedAtomicPtr::<_, 3>::new(boxed, 1usize);

    assert_eq!(ptr.tag(Ordering::Relaxed), 1);
    assert_eq!(ptr.ptr(Ordering::Relaxed).as_ptr(), boxed);
    ptr.set_tag(3, Ordering::Release, Ordering::Relaxed);
    assert_eq!(ptr.tag(Ordering::Relaxed), 3);
    assert_eq!(ptr.ptr(Ordering::Relaxed).as_ptr(), boxed);
    let old = ptr.compare_exchange_weak_tag(3, 4, Ordering::Release, Ordering::Relaxed);
    assert_eq!(old, Ok(3));
    assert_eq!(ptr.tag(Ordering::Relaxed), 4);
    assert_eq!(ptr.ptr(Ordering::Relaxed).as_ptr(), boxed);
}

#[test]
fn tagged_ptr_exchange_failure() {
    let boxed = Box::into_raw(Box::new(42usize));
    let ptr = TaggedAtomicPtr::<_, 3>::new(boxed, 1usize);

    assert_eq!(ptr.tag(Ordering::Relaxed), 1);
    assert_eq!(ptr.ptr(Ordering::Relaxed).as_ptr(), boxed);

    let old = ptr.compare_exchange_weak_tag(3, 4, Ordering::Release, Ordering::Relaxed);
    assert_eq!(old, Err(1));
    assert_eq!(ptr.tag(Ordering::Relaxed), 1);
    assert_eq!(ptr.ptr(Ordering::Relaxed).as_ptr(), boxed);
}

#[test]
fn value_inline() {
    let val = Value::new(3usize);

    let inner = val.get();
    assert_eq!(inner, 3usize);
}

#[test]
fn value_inline_struct() {
    #[derive(Default, PartialEq, Debug)]
    struct Small {
        c: f32,
    }
    let val = Value::new(Small { c: 3.2 });

    let inner = val.get();
    assert_eq!(inner.c, 3.2);
}

#[test]
fn value_boxed() {
    #[derive(Default, PartialEq, Debug)]
    struct Big {
        a: usize,
        b: f32,
        c: u32,
    }
    let val = Value::new(Big {
        a: 42,
        b: 2.25,
        c: 7,
    });

    let inner = val.get();
    assert_eq!(
        inner,
        Big {
            a: 42,
            b: 2.25,
            c: 7
        }
    );
}

#[test]
fn value_inline_drop() {
    #[derive(PartialEq, Debug)]
    #[repr(transparent)]
    struct Small<'a> {
        inner: &'a mut usize,
    }
    impl Drop for Small<'_> {
        fn drop(&mut self) {
            *self.inner += 1;
        }
    }
    let mut dropped = 0;
    {
        let inner = {
            let val = Value::new(Small {
                inner: &mut dropped,
            });

            val.get()
        };
        assert_eq!(*inner.inner, 0);
    }
    assert_eq!(dropped, 1);
    {
        let _val = Value::new(Small {
            inner: &mut dropped,
        });
    }
    assert_eq!(dropped, 2);
}

#[test]
fn value_boxed_drop() {
    #[derive(PartialEq, Debug)]
    struct Big<'a> {
        inner: &'a mut usize,
        pad: [usize; 3],
    }
    impl Drop for Big<'_> {
        fn drop(&mut self) {
            *self.inner += 1;
        }
    }
    let mut dropped = 0;
    {
        let inner = {
            let val = Value::new(Big {
                inner: &mut dropped,
                pad: [0; 3],
            });

            val.get()
        };
        assert_eq!(*inner.inner, 0);
    }
    assert_eq!(dropped, 1);
    {
        let _val = Value::new(Big {
            inner: &mut dropped,
            pad: [0; 3],
        });
    }
    assert_eq!(dropped, 2);
}
