use core::{
    cell::Cell,
    ops::{Deref, DerefMut},
    ptr::NonNull,
};

#[derive(Debug, PartialEq, Eq, Hash, PartialOrd, Ord, Clone, Copy)]
#[repr(transparent)]
pub struct SendPtr<T>(NonNull<T>);

impl<T> SendPtr<T> {
    pub fn as_ptr(&self) -> *mut T {
        self.0.as_ptr()
    }
    pub unsafe fn new_unchecked(t: *const T) -> Self {
        unsafe { Self(NonNull::new_unchecked(t.cast_mut())) }
    }
    pub fn new(t: *const T) -> Option<Self> {
        NonNull::new(t.cast_mut()).map(Self)
    }
    pub fn cast<U>(self) -> SendPtr<U> {
        SendPtr(self.0.cast::<U>())
    }
}

impl<T> Deref for SendPtr<T> {
    type Target = T;

    fn deref(&self) -> &Self::Target {
        unsafe { &*self.0.as_ptr() }
    }
}

impl<T> DerefMut for SendPtr<T> {
    fn deref_mut(&mut self) -> &mut Self::Target {
        unsafe { &mut *self.0.as_ptr() }
    }
}

pub struct XorShift64Star {
    state: Cell<u64>,
}

impl XorShift64Star {
    /// Initializes the prng with a seed. Provided seed must be nonzero.
    pub fn new(seed: u64) -> Self {
        XorShift64Star {
            state: Cell::new(seed),
        }
    }

    /// Returns a pseudorandom number.
    pub fn next(&self) -> u64 {
        let mut x = self.state.get();
        debug_assert_ne!(x, 0);
        x ^= x >> 12;
        x ^= x << 25;
        x ^= x >> 27;
        self.state.set(x);
        x.wrapping_mul(0x2545_f491_4f6c_dd1d)
    }

    /// Return a pseudorandom number from `0..n`.
    pub fn next_usize(&self, n: usize) -> usize {
        (self.next() % n as u64) as usize
    }
}
